# -*- coding: utf-8 -*-

from pyqtgraph.Qt.QtCore import QEvent, QObject, pyqtSignal
from pyqtgraph.Qt.QtGui import QMouseEvent
from pyqtgraph.Qt.QtWidgets import QWidget


__all__ = ['clickable']


def clickable(widget: QWidget):
    '''Adds a clicked() signal to a widget such as QLineEdit that
    ordinarily does not provide notifications of clicks.

    Example:
    -------
    myWidget = QWidget()
    widgetClicked = clickable(myWidget)
    widgetClicked.connect(lambda: print("clicked"))

    Explanation:
    -----------
    After the call to clickable(), myWidget now processes
    MouseButtonRelease events by emitting the clicked() signal.
    clickable() returns the signal itself for convenience.
    '''

    class Filter(QObject):

        clicked = pyqtSignal()

        def eventFilter(self, obj: QObject, event: QMouseEvent) -> bool:
            if event.type() == QEvent.Type.MouseButtonRelease:
                pos = (event.position().toPoint()
                       if hasattr(event, 'position') else event.pos())
                if obj.rect().contains(pos):
                    self.clicked.emit()
                    return True
            return super().eventFilter(obj, event)

    event_filter = Filter(widget)
    widget.installEventFilter(event_filter)
    return event_filter.clicked



from QVideo.lib import QVideoSource, QFilterBank
from QVideo.lib.types import Image
from pyqtgraph.Qt import QtCore
import numpy as np
from pyqtgraph import GraphicsLayoutWidget, ImageItem
import logging


logger = logging.getLogger(__name__)

__all__ = ['QVideoScreen']


class QVideoScreen(GraphicsLayoutWidget):

    '''Video display widget.

    Displays frames from a :class:`~QVideo.lib.QVideoSource` in real time,
    with optional frame-rate throttling and image filtering via
    :class:`~QVideo.lib.QFilterBank`.

    Inherits from :class:`pyqtgraph.GraphicsLayoutWidget`.

    Parameters
    ----------
    size : tuple[int, int]
        Initial widget dimensions in pixels. Default: ``(640, 480)``.
    framerate : int | None
        Maximum display frame rate in frames per second.
        ``None`` means no throttling. Default: ``None``.
    *args :
        Forwarded to :class:`pyqtgraph.GraphicsLayoutWidget`.
    **kwargs :
        Forwarded to :class:`pyqtgraph.GraphicsLayoutWidget`.

    Properties
    ----------
    framerate : int | None
        Maximum display frame rate [fps]. Setting this updates the throttle
        interval.
    source : QVideoSource
        The video source. Setting this connects :attr:`newFrame` and
        :attr:`shapeChanged` to the display.
    '''

    def __init__(self, *args,
                 size: tuple[int, int] = (640, 480),
                 framerate: int | None = None,
                 **kwargs) -> None:
        super().__init__(*args, size=size, **kwargs)
        self.framerate = framerate
        self._ready = True
        self._pending = None
        self._setupUi()
        self._timer = QtCore.QTimer()
        self._source = None
        self.filter = QFilterBank(self)
        self.filter.setVisible(False)

    def _setupUi(self) -> None:
        self.ci.layout.setContentsMargins(0, 0, 0, 0)
        self.view = self.addViewBox(enableMenu=False,
                                    enableMouse=False)
        self.view.invertY(True)
        self.view.setAspectLocked(True)
        self.view.setDefaultPadding(0)
        self.updateShape(self.size())
        self.image = ImageItem(axisOrder='row-major')
        self.view.addItem(self.image)

    @property
    def framerate(self) -> int | None:
        '''Maximum display frame rate [fps], or ``None`` for no throttling.'''
        return self._framerate

    @framerate.setter
    def framerate(self, framerate: int | None) -> None:
        if framerate is not None and framerate <= 0:
            raise ValueError(f'framerate must be positive, got {framerate}')
        self._framerate = framerate
        self._interval = 0 if framerate is None else int(1000 / framerate)

    def _setready(self) -> None:
        self._ready = True
        if self._pending is not None:
            self.setImage(self._pending)

    @property
    def source(self) -> QVideoSource:
        '''The video source providing frames to display.'''
        return self._source

    @source.setter
    def source(self, source: QVideoSource) -> None:
        if self._source is not None:
            self._source.shapeChanged.disconnect(self.updateShape)
            self._source.newFrame.disconnect(self.setImage)
        self._source = source
        self.updateShape(source.shape)
        self._source.shapeChanged.connect(self.updateShape)
        self._source.newFrame.connect(self.setImage)

    @QtCore.pyqtSlot(np.ndarray)
    def setImage(self, image: Image) -> None:
        '''Display a new video frame, subject to frame-rate throttling.

        Passes the frame through :attr:`filter` before display.  If the
        throttle interval has not yet elapsed, the frame is buffered as the
        most recent pending frame; when the interval expires the buffered
        frame is displayed immediately so no extra latency accumulates.

        Parameters
        ----------
        image : Image
            The frame to display.
        '''
        if self._ready:
            self.image.setImage(self.filter(image), autoLevels=False)
            self._ready = False
            self._pending = None
            self._timer.singleShot(self._interval, self._setready)
        else:
            self._pending = image

    def sizeHint(self) -> QtCore.QSize:
        '''Return the source's natural frame size as the preferred widget size.'''
        if self.source is not None:
            return self.source.shape
        return super().sizeHint()

    def hasHeightForWidth(self) -> bool:
        '''Return True when the source frame size is known.'''
        return self.source is not None

    def heightForWidth(self, width: int) -> int:
        '''Return the height that preserves the source aspect ratio.'''
        if self.source is not None:
            shape = self.source.shape
            return width * shape.height() // shape.width()
        return super().heightForWidth(width)

    @QtCore.pyqtSlot(QtCore.QSize)
    def updateShape(self, shape: QtCore.QSize) -> None:
        '''Resize the display to match the video frame dimensions.

        Parameters
        ----------
        shape : QtCore.QSize
            New frame dimensions.
        '''
        logger.debug(f'Resizing to {shape}')
        self.view.setRange(xRange=(0, shape.width()),
                           yRange=(0, shape.height()),
                           padding=0, update=True)
        self.setMinimumSize(shape / 2)
        self.updateGeometry()

    @classmethod
    def example(cls: 'QVideoScreen') -> None:  # pragma: no cover
        '''Demonstrate the video screen with a noise source.'''
        import pyqtgraph as pg
        from QVideo.cameras.Noise import QNoiseSource

        app = pg.mkQApp()
        screen = cls()
        source = QNoiseSource(blacklevel=48, whitelevel=128)
        screen.source = source.start()
        screen.show()
        pg.exec()


if __name__ == '__main__':  # pragma: no cover
    QVideoScreen.example()



from .types import Image
from .clickable import clickable
from .QCamera import QCamera
from .QVideoSource import QVideoSource
from .QCameraTree import QCameraTree
from .QFilterBank import QFilterBank
from .QVideoReader import QVideoReader
from .QVideoWriter import QVideoWriter
from .QVideoScreen import QVideoScreen
from .chooser import choose_camera
from .QListCameras import QListCameras
from .QFPSMeter import QFPSMeter
from .VideoFilter import VideoFilter


__all__ = '''Image
clickable choose_camera QListCameras
QCamera QVideoSource QCameraTree QFilterBank
QVideoReader QVideoWriter QVideoScreen
QFPSMeter VideoFilter'''.split()



from QVideo.lib import QCameraTree
from QVideo.lib.QCameraTree import Source
from QVideo.cameras.Genicam import QGenicamCamera
from genicam.genapi import (IValue, EAccessMode, EVisibility,
                            ICategory, ICommand, IEnumeration,
                            IBoolean, IInteger, IFloat, IString)
import logging


logger = logging.getLogger(__name__)


__all__ = ['QGenicamTree']


class QGenicamTree(QCameraTree):

    '''Camera property tree for :class:`~QVideo.cameras.Genicam.QGenicamCamera`.

    Builds a :class:`~QVideo.lib.QCameraTree.QCameraTree` from the camera's
    GenICam node map and exposes visibility and per-feature enable/disable
    controls.

    Parameters
    ----------
    camera : QGenicamCamera
        Camera instance to use.
    visibility : EVisibility
        Maximum GenICam visibility level to display.
        Default: ``EVisibility.Guru``.
    controls : list of str or None
        If given, only nodes whose names appear in this list are shown;
        all others are hidden.  Default: ``None`` (show all).
    *args :
        Forwarded to :class:`~QVideo.lib.QCameraTree.QCameraTree`.
    **kwargs :
        Forwarded to :class:`~QVideo.lib.QCameraTree.QCameraTree`.
    '''

    def __init__(self, *args,
                 camera: Source,
                 visibility: EVisibility = EVisibility.Guru,
                 controls: list[str] | None = None,
                 **kwargs) -> None:
        description = self.description(camera)
        super().__init__(camera, description, *args, **kwargs)
        self._visibility = visibility
        self.controls = controls
        self.visibility = visibility
        self._updateEnabled()

    def description(self, camera: QGenicamCamera) -> dict:
        '''Return a dictionary describing the node map of the camera'''
        root = camera.node('Root')
        description = self.describe(root)
        return description['children']

    def describe(self, feature: IValue) -> dict[str, object]:
        '''Return a dictionary describing the specified feature'''
        this = dict(name=feature.node.name,
                    title=feature.node.display_name,
                    visibility=feature.node.visibility)
        mode = feature.node.get_access_mode()
        if mode == EAccessMode.NI:
            return this
        if isinstance(feature, ICategory):
            this['type'] = 'group'
            this['children'] = [self.describe(f)
                                for f in feature.features]
            return this
        if isinstance(feature, ICommand):
            this['type'] = 'action'
            return this
        if mode not in (EAccessMode.RW, EAccessMode.RO):
            return this
        if isinstance(feature, IEnumeration):
            this['type'] = 'list'
            this['value'] = this['default'] = feature.to_string()
            this['limits'] = [v.symbolic for v in feature.entries]
        elif isinstance(feature, IBoolean):
            this['type'] = 'bool'
            this['value'] = this['default'] = feature.value
        elif isinstance(feature, IInteger):
            this['type'] = 'int'
            this['value'] = this['default'] = feature.value
            this['min'] = feature.min
            this['max'] = feature.max
            this['step'] = feature.inc
        elif isinstance(feature, IFloat):
            this['type'] = 'float'
            this['value'] = this['default'] = feature.value
            this['min'] = feature.min
            this['max'] = feature.max
            this['units'] = feature.unit
            if feature.has_inc():
                this['step'] = feature.inc
        elif isinstance(feature, IString):
            this['type'] = 'str'
            this['value'] = this['default'] = feature.value
        else:
            # FIXME: Support for IRegister nodes
            logger.debug(
                f'Unsupported node type: {feature.node.name}: {type(feature)}')
        return this

    def _connectSignals(self) -> None:
        super()._connectSignals()
        for item in self.listAllItems():
            p = item.param
            p.sigValueChanged.connect(self._handleItemChanges)

    def _handleItemChanges(self) -> None:
        logger.debug('Handling item changes')
        self._updateVisible()
        self._updateEnabled()
        self._updateLimits()
        self._updateValues()

    def _updateVisible(self) -> None:
        for item in self.listAllItems()[1:]:
            p = item.param
            p.setOpts(visible=self.visible(p))

    def visible(self, param) -> bool:
        ptype = param.opts['type']
        if ptype in ('action', None):
            return False
        if ptype == 'group':
            return any(self.visible(c) for c in param.children())
        visibility = param.opts.get('visibility', EVisibility.Invisible)
        return visibility <= self.visibility

    def _updateEnabled(self) -> None:
        for item in self.listAllItems()[1:]:
            p = item.param
            if p.opts.get('visible', False):
                name = p.opts['name']
                if self.camera.has_node(name):
                    p.setOpts(enabled=self.camera.is_readwrite(name))

    def _updateValues(self) -> None:
        '''Refresh read-only Parameter values from the live GenICam node map.

        Called after every property change so that derived read-only nodes
        (e.g. ``AcquisitionResultingFrameRate``) reflect the current hardware
        state in the UI.  Only visible read-only leaf parameters are updated.
        Signals are blocked during the update to prevent re-entrant calls.
        '''
        for item in self.listAllItems()[1:]:
            p = item.param
            if not p.opts.get('visible', False):
                continue
            name = p.opts.get('name')
            if not self.camera.has_node(name):
                continue
            node = self.camera.node(name)
            if node.node.get_access_mode() != EAccessMode.RO:
                continue
            if isinstance(node, IEnumeration):
                value = node.to_string()
            elif isinstance(node, (IBoolean, IInteger, IFloat, IString)):
                value = node.value
            else:
                continue
            p.blockSignals(True)
            try:
                p.setValue(value)
            finally:
                p.blockSignals(False)

    def _updateLimits(self) -> None:
        '''Refresh Parameter constraints from the live GenICam node values.

        Called after every property change so that dependent nodes (e.g.
        ``OffsetX`` range after a ``Width`` change) reflect the current
        hardware state in the UI.  Only visible leaf parameters are updated.
        '''
        for item in self.listAllItems()[1:]:
            p = item.param
            if not p.opts.get('visible', False):
                continue
            name = p.opts.get('name')
            if not self.camera.has_node(name):
                continue
            node = self.camera.node(name)
            if isinstance(node, IInteger):
                p.setOpts(limits=(node.min, node.max), step=node.inc)
            elif isinstance(node, IFloat):
                opts = {'limits': (node.min, node.max)}
                if node.has_inc():
                    opts['step'] = node.inc
                p.setOpts(**opts)
            elif isinstance(node, IEnumeration):
                p.setOpts(limits=[v.symbolic for v in node.entries])

    @property
    def controls(self) -> list[str] | None:
        return self._controls

    @controls.setter
    def controls(self, controls: list[str] | None) -> None:
        self._controls = controls
        for item in self.listAllItems()[1:]:
            p = item.param
            name = p.opts['name']
            node = self.camera.node(name)
            visible = controls is None or name in controls
            p.opts['visibility'] = (node.node.visibility
                                    if node is not None and visible
                                    else EVisibility.Invisible)
        self._updateVisible()

    @property
    def visibility(self) -> EVisibility:
        return self._visibility

    @visibility.setter
    def visibility(self, visibility: EVisibility) -> None:
        self._visibility = visibility
        self._updateVisible()


if __name__ == '__main__':  # pragma: no cover
    QGenicamTree.example()
